/*	$NetBSD$	*/

/*-
 * Copyright (c) 2010 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/device.h>

#include <uvm/uvm_extern.h>

#include <arm/imx/imx51reg.h>
#include <arm/imx/imx51var.h>
#include <arm/imx/imx51_gpio.h>
#include <arm/imx/imx51_iomux.h>
#include <arm/imx/imx51_iomuxreg.h>

#include "locators.h"

struct axi_softc {
	device_t sc_dev;
	bus_space_tag_t sc_iot;
	bus_dma_tag_t sc_dmat;
	struct imx51gpio_softc *sc_gpio[4];
	struct imx51iomux_softc *sc_iomux;
};

static int axi_match(device_t, cfdata_t, void *);
static void axi_attach(device_t, device_t, void *);
static int axi_critical_search(device_t, struct cfdata *, const int *, void *);
static int axi_search(device_t, cfdata_t, const int *, void *);
static int axi_print(void *, const char *);

CFATTACH_DECL_NEW(axi, sizeof(struct axi_softc),
    axi_match, axi_attach, NULL, NULL);

/* ARGSUSED */
static int
axi_match(device_t parent __unused, cfdata_t match __unused,
    void *aux __unused)
{
	return 1;
}

/* ARGSUSED */
static void
axi_attach(device_t parent __unused, device_t self, void *aux __unused)
{
	struct axi_softc *sc;
	struct axi_attach_args aa;

	sc = device_private(self);
	sc->sc_iot = &imx51_bs_tag;
	sc->sc_dmat = &imx51_bus_dma_tag;

	aprint_normal(": AMBA AXI interface\n");
	aprint_naive("\n");

	/*
	 * Attach all devices
	 */
	config_search_ia(axi_critical_search, self, "axi", &aa);
	config_search_ia(axi_search, self, "axi", &aa);


#if 1
	/* IOMUX for UART1 */
	AXI_IOMUXC_MUXPAD(sc, UART1_RXD, ALT0,
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_SR_FAST |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_DS_HIGH |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PUE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PKE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_HYS);
	axi_iomux_set(sc,
	    IMX51_IOMUXC_UART1_IPP_UART_RXD_MUX_SELECT_INPUT_REG,
	    IMX51_IOMUXC_DAISY0);

	AXI_IOMUXC_MUXPAD(sc, UART1_TXD, ALT0,
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_SR_FAST |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_DS_HIGH |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PUE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PKE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_HYS);

	AXI_IOMUXC_MUXPAD(sc, UART1_RTS, ALT0,
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_DS_HIGH |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PUE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PKE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_HYS);
	axi_iomux_set(sc,
	    IMX51_IOMUXC_UART1_IPP_UART_RTS_B_SELECT_INPUT_REG,
	    IMX51_IOMUXC_DAISY0);

	AXI_IOMUXC_MUXPAD(sc, UART1_CTS, ALT0,
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_DS_HIGH |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PUE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_PKE |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_HYS);
#endif
}

/* ARGSUSED */
static int
axi_critical_search(device_t parent, struct cfdata *cf, const int *ldesc __unused, void *aux)
{
	device_t d;
	struct axi_softc *sc;
	struct axi_attach_args *aa;

	sc = device_private(parent);
	aa = aux;

	if ((strcmp(cf->cf_name, "tzic") != 0) &&
	    (strcmp(cf->cf_name, "imx51gpio") != 0) &&
	    (strcmp(cf->cf_name, "imx51iomux") != 0))
		return 0;

	aa->aa_iot = sc->sc_iot;
	aa->aa_dmat = sc->sc_dmat;
	aa->aa_addr = cf->cf_loc[AXICF_ADDR];
	aa->aa_irq = cf->cf_loc[AXICF_IRQ];
	aa->aa_nirq = cf->cf_loc[AXICF_NIRQ];
	aa->aa_size = 0;

	if (config_match(parent, cf, aux) > 0) {
		d = config_attach(parent, cf, aux, axi_print);

		/* save iomux softc for axi_iomux_*() functions */
		if ((strcmp(cf->cf_name, "imx51iomux") == 0) &&
		    (aa->aa_addr == IMX51_IOMUX_BASE))
			sc->sc_iomux = device_private((device_t)d);

		/* save gpio softc for axi_gpio_*() functions */
		if (strcmp(cf->cf_name, "imx51gpio") == 0) {
			if (aa->aa_addr == IMX51_GPIO1_BASE)
				sc->sc_gpio[0] = device_private((device_t)d);
			if (aa->aa_addr == IMX51_GPIO2_BASE)
				sc->sc_gpio[1] = device_private((device_t)d);
			if (aa->aa_addr == IMX51_GPIO3_BASE)
				sc->sc_gpio[2] = device_private((device_t)d);
			if (aa->aa_addr == IMX51_GPIO4_BASE)
				sc->sc_gpio[3] = device_private((device_t)d);
		}
	}
	return 0;
}

/* ARGSUSED */
static int
axi_search(device_t parent, cfdata_t cf, const int *ldesc __unused,
    void *aux)
{
	struct axi_softc *sc;
	struct axi_attach_args *aa;

	sc = device_private(parent);
	aa = aux;

	aa->aa_iot = sc->sc_iot;
	aa->aa_dmat = sc->sc_dmat;
	aa->aa_addr = cf->cf_loc[AXICF_ADDR];
	aa->aa_irq = cf->cf_loc[AXICF_IRQ];
	aa->aa_nirq = cf->cf_loc[AXICF_NIRQ];
	aa->aa_size = 0;

	if (config_match(parent, cf, aux) > 0)
		config_attach(parent, cf, aux, axi_print);

	return 0;
}

/* ARGSUSED */
static int
axi_print(void *aux, const char *name __unused)
{
	struct axi_attach_args *aa;

	aa = (struct axi_attach_args *)aux;

	if (aa->aa_size)
		aprint_normal(" addr 0x%lx", aa->aa_addr);
	if (aa->aa_size > 1)
		aprint_normal("-0x%lx", aa->aa_addr + aa->aa_size - 1);
	if (aa->aa_irq >= 0) {
		aprint_normal(" irq %d", aa->aa_irq);
		if (aa->aa_nirq >= 0) {
			aprint_normal("-%d", aa->aa_irq + aa->aa_nirq - 1);
		}
	}

	return UNCONF;
}

void
axi_gpio_dir(struct axi_softc *sc, axi_gpioclass_t gpioclass, int pin, int flags)
{
	struct imx51gpio_softc *gpio;

	gpio = sc->sc_gpio[gpioclass];
	if (gpio == NULL)
		return;

	imx51gpio_pin_ctl(gpio, pin, flags);
}

void
axi_gpio_write(struct axi_softc *sc, axi_gpioclass_t gpioclass, int pin, int val)
{
	struct imx51gpio_softc *gpio;

	gpio = sc->sc_gpio[gpioclass];
	if (gpio == NULL)
		return;

	imx51gpio_pin_write(gpio, pin, val);
}

void
axi_iomux_set(struct axi_softc *sc, uint32_t reg, uint32_t val)
{
	if (sc->sc_iomux == NULL)
		return;

	imx51iomux_write(sc->sc_iomux, reg, val);
}
