/*	$NetBSD$	*/

/*-
 * Copyright (c) 2010 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/timetc.h>
#include <sys/device.h>

#include <machine/bus.h>

#include <arm/imx/imx51reg.h>
#include <arm/imx/imx51var.h>

struct epit_softc {
	device_t sc_dev;
	bus_addr_t sc_addr;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	void *sc_ih;
	bus_size_t sc_size;
};

static int epit_match(device_t, struct cfdata *, void *);
static void epit_attach(device_t, device_t, void *);
static void epit_init(struct epit_softc *);
static int epit_timer_intr(struct epit_softc *);

CFATTACH_DECL_NEW(imx51epit, sizeof(struct epit_softc),
    epit_match, epit_attach, NULL, NULL);

/* ARGSUSED */
static int
epit_match(device_t parent __unused, struct cfdata *match __unused,
    void *aux)
{
	struct axi_attach_args *aa;

	aa = aux;

	switch (aa->aa_addr) {
	case IMX51_EPIT1_BASE:
	case IMX51_EPIT2_BASE:
		break;
	default:
		return 0;
	}

	aa->aa_size = IMX51_EPIT_SIZE;

	return 1;
}

/* ARGSUSED */
static void
epit_attach(device_t parent __unused, device_t self, void *aux)
{
	struct epit_softc *sc;
	struct axi_attach_args *aa;

	aprint_normal(": Enhanced Periodic Interrupt Timer\n");
	aprint_naive("\n");

	aa = aux;
	sc = device_private(self);
	sc->sc_dev = self;
	sc->sc_iot = aa->aa_iot;
	sc->sc_addr = aa->aa_addr;
	sc->sc_size = aa->aa_size;

	if (bus_space_map(sc->sc_iot, sc->sc_addr, aa->aa_size, 0, &sc->sc_ioh))
	    panic("%s: Cannot map registers", device_xname(self));

	epit_init(sc);

	sc->sc_ih = intr_establish(IRQ_EPIT1, IPL_CLOCK, 0,
	    epit_timer_intr, sc);
	if (sc->sc_ih == NULL)
		aprint_error_dev(self, "unable to register interrupt\n");
}

static void
epit_init(struct epit_softc *sc)
{
	uint32_t epitctrl;

	/*
	 * reset timer counter
	 */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_CONTROL_REG, 0);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_CONTROL_REG, 0);

	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_LOAD_REG, 0x00009876);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_COMPARE_REG, 0x00004321);

	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_CONTROL_REG,
	    IMX51_EPIT_CONTROL_CLKSRC(2));
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_STATUS_REG, IMX51_EPIT_STATUS_OCIF);	/* clear */

	epitctrl = 
	    IMX51_EPIT_CONTROL_CLKSRC(2) |
	    IMX51_EPIT_CONTROL_OM(3) |
	    IMX51_EPIT_CONTROL_IOVW |
	    IMX51_EPIT_CONTROL_PRESCALAR(0) |
	    IMX51_EPIT_CONTROL_RLD |
	    IMX51_EPIT_CONTROL_OCIEN |
	    IMX51_EPIT_CONTROL_ENMOD;
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_CONTROL_REG,
	    epitctrl);

	/* enable timer */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_EPIT_CONTROL_REG,
	    epitctrl | IMX51_EPIT_CONTROL_EN);
}

static int
epit_timer_intr(void *arg)
{
	struct epit_softc *sc;

	sc = (struct epit_softc *)arg;

	/* XXX: notyet */
	printf("%s:%d\n", __func__, __LINE__);

	return 1;
}
