/*	$NetBSD$	*/

/*-
 * Copyright (c) 2010 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/timetc.h>
#include <sys/device.h>

#include <machine/bus.h>

#include <arm/imx/imx51reg.h>
#include <arm/imx/imx51var.h>

/* specified stathz. if set 0, statclock shared with hardclock */
#ifndef STATHZ
#define STATHZ	64
#endif

struct gpt_softc {
	device_t sc_dev;
	bus_addr_t sc_addr;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	bus_size_t sc_size;
	void *sc_ih;
	int sc_irq;
	uint32_t sc_timer1_clk_per_hz;
	uint32_t sc_timer2_clk_per_hz;
};

static struct gpt_softc *clock_sc;

static int gpt_match(device_t, struct cfdata *, void *);
static void gpt_attach(device_t, device_t, void *);
static u_int gpt_get_timecount(struct timecounter *);
static void gpt_capture1_set(struct gpt_softc *, uint32_t);
static void gpt_capture2_set(struct gpt_softc *, uint32_t);
static int gpt_timer_intr(void *);	/* interrupt handler */

static struct timecounter gpt_timecounter = {
	.tc_get_timecount = gpt_get_timecount,
	.tc_poll_pps = NULL,
	.tc_counter_mask = ~0U,
	.tc_frequency = IMX51_IPG_CLK_HIGHFREQ,
	.tc_name = "imx51gpt",
	.tc_quality = 100,
	.tc_priv = NULL,
	.tc_next = NULL
};

CFATTACH_DECL_NEW(imx51gpt, sizeof(struct gpt_softc),
    gpt_match, gpt_attach, NULL, NULL);

/* ARGSUSED */
static int
gpt_match(device_t parent __unused, struct cfdata *match __unused,
    void *aux)
{
	struct axi_attach_args *aa;

	aa = aux;

	switch (aa->aa_addr) {
	case IMX51_GPT_BASE:
		break;
	default:
		return 0;
	}

	aa->aa_size = IMX51_GPT_SIZE;

	return 1;
}

/* ARGSUSED */
static void
gpt_attach(device_t parent __unused, device_t self, void *aux)
{
	struct gpt_softc *sc;
	struct axi_attach_args *aa;

	aprint_normal(": General Purpose Timer\n");
	aprint_naive("\n");

	aa = aux;
	sc = device_private(self);
	sc->sc_dev = self;
	sc->sc_iot = aa->aa_iot;
	sc->sc_addr = aa->aa_addr;
	sc->sc_size = aa->aa_size;
	sc->sc_irq = aa->aa_irq;

	if (bus_space_map(sc->sc_iot, sc->sc_addr, aa->aa_size, 0, &sc->sc_ioh))
	    panic("%s: Cannot map registers", device_xname(self));

	/*
	 * clock_sc would be referenced from cpu_initclocks(),
	 * setstatclockrate(), interrupt routines, etc.
	 */
	clock_sc = sc;	/* XXX: save softc of gpt0 */

	/* disable and clear all interrupts */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTERRUPT_REG, 0);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTSTATUS_REG,
	    0xffffffff);

	/*
	 * set interrupts handler
	 */
	sc->sc_ih = intr_establish(sc->sc_irq, IPL_CLOCK, 0,
	    gpt_timer_intr, NULL);
	if (sc->sc_ih == NULL)
		panic("gpt_attach: unable to register interrupt\n");

	/* start timer */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_CONTROL_REG,
	    IMX51_GPT_CONTROL_FRR |
	    IMX51_GPT_CONTROL_CLKSRC(2) | /* 2 is IPG_CLK_HIGHFREQ */
	    IMX51_GPT_CONTROL_WAITEN |
	    IMX51_GPT_CONTROL_EN);
}

void
gpt_capture1_set(struct gpt_softc *sc, uint32_t clk)
{
	uint32_t v;

	sc->sc_timer1_clk_per_hz = clk;

	v = bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COUNTER_REG);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COMPARE1_REG,
	    v + clk);

	/* enable interrupts */
	v = bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTERRUPT_REG);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTERRUPT_REG,
	    v | IMX51_GPT_INTERRUPT_OF1);
}

void
gpt_capture2_set(struct gpt_softc *sc, uint32_t clk)
{
	uint32_t v;

	sc->sc_timer2_clk_per_hz = clk;

	v = bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COUNTER_REG);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COMPARE2_REG,
	    v + clk);

	/* enable interrupts */
	v = bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTERRUPT_REG);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTERRUPT_REG,
	    v | IMX51_GPT_INTERRUPT_OF2);
}

static int
gpt_timer_intr(void *arg)
{
	struct gpt_softc *sc;
	uint32_t count, status;

	sc = clock_sc;

	/* read and clear interrupt status */
	status = bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTSTATUS_REG);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_INTSTATUS_REG,
	    0xffffffff);

	count = bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COUNTER_REG);

	if (status & IMX51_GPT_INTERRUPT_OF1) {
//		printf("%s:%d: HARDCLOCK: cnt=%d\n", __func__, __LINE__, count);
		bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COMPARE1_REG,
		    count + sc->sc_timer1_clk_per_hz);
		hardclock(arg);
	}

	if (status & IMX51_GPT_INTERRUPT_OF2) {
//		printf("%s:%d: STATCLOCK: cnt=%d\n", __func__, __LINE__, count);
		bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COMPARE2_REG,
		    count + sc->sc_timer2_clk_per_hz);
		statclock(arg);
	}

	return 1;
}

/* ARGSUSED */
static u_int
gpt_get_timecount(struct timecounter *tc __unused)
{
	struct gpt_softc *sc;

	sc = clock_sc;

	return bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_GPT_COUNTER_REG);
}

void
cpu_initclocks(void)
{
	struct gpt_softc *sc;

	sc = clock_sc;
	if (sc == NULL)
		panic("cpu_initclocks: gpt not attached");

	profhz = stathz = STATHZ;

	/*
	 * set capture counters for hardclock and statclock
	 */
	gpt_capture1_set(sc, IMX51_IPG_CLK_HIGHFREQ / hz);
	gpt_capture2_set(sc, IMX51_IPG_CLK_HIGHFREQ / stathz);

	tc_init(&gpt_timecounter);
}

void
setstatclockrate(int newhz)
{
	struct gpt_softc *sc;

	sc = clock_sc;
	gpt_capture2_set(sc, IMX51_IPG_CLK_HIGHFREQ / newhz);
}
