/*	$NetBSD$	*/

/*-
 * Copyright (c) 2010 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/timetc.h>
#include <sys/device.h>

#include <machine/bus.h>

#include <arm/imx/imx51reg.h>
#include <arm/imx/imx51var.h>
#include <arm/imx/imx51_iomuxreg.h>

struct imx51pwm_softc {
	device_t sc_dev;
	struct axi_softc *sc_axi;
	bus_addr_t sc_addr;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	void *sc_ih;
};

static int imx51pwm_match(device_t, struct cfdata *, void *);
static void imx51pwm_attach(device_t, device_t, void *);
static void imx51pwm_iomux_init(struct imx51pwm_softc *);
static void imx51pwm_init(struct imx51pwm_softc *);
static void imx51pwm_rate(struct imx51pwm_softc *, int);
static int imx51pwm_intr(void *);	/* interrupt handler */

CFATTACH_DECL_NEW(imx51pwm, sizeof(struct imx51pwm_softc),
    imx51pwm_match, imx51pwm_attach, NULL, NULL);

/* ARGSUSED */
static int
imx51pwm_match(device_t parent __unused, struct cfdata *match __unused,
    void *aux)
{
	struct axi_attach_args *aa;

	aa = aux;

	switch (aa->aa_addr) {
	case IMX51_PWM1_BASE:
	case IMX51_PWM2_BASE:
		break;
	default:
		return 0;
	}

	aa->aa_size = IMX51_PWM_SIZE;

	return 1;
}

static void
imx51pwm_attach(device_t parent, device_t self, void *aux)
{
	struct imx51pwm_softc *sc;
	struct axi_attach_args *aa;

	aprint_normal(": Pulse-Width Modulator\n");
	aprint_naive("\n");

	aa = aux;
	sc = device_private(self);
	sc->sc_axi = device_private(parent);
	sc->sc_dev = self;
	sc->sc_iot = aa->aa_iot;
	sc->sc_addr = aa->aa_addr;

	if (bus_space_map(sc->sc_iot, sc->sc_addr, aa->aa_size, 0, &sc->sc_ioh))
	    panic("%s: Cannot map registers", device_xname(self));

	/*
	 * set interrupts handler
	 */
	sc->sc_ih = intr_establish(aa->aa_irq, IPL_CLOCK, 0,
	    imx51pwm_intr, sc);
	if (sc->sc_ih == NULL)
		panic("imx51pwm_attach: unable to register interrupt\n");

	imx51pwm_iomux_init(sc);
	imx51pwm_init(sc);

	imx51pwm_rate(sc, 255);	/* LCD brightness on NetWalker */

#if 0 /* interrupt is not used */
	/* enable interrupts */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_INTERRUPT_REG,
	    IMX51_PWM_INTERRUPT_RIE);
#endif
}

static void
imx51pwm_iomux_init(struct imx51pwm_softc *sc)
{
#ifdef NETWALKER
	AXI_IOMUXC_MUXPAD(sc->sc_axi, GPIO1_2, ALT1,
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_DS_HIGH |
	    IMX51_IOMUXC_SW_PAD_CTL_PAD_OD_ENABLE);
#endif /* NETWALKER */
}

static void
imx51pwm_init(struct imx51pwm_softc *sc)
{
	uint32_t val;

	/* disable interrupts */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_INTERRUPT_REG, 0);

	/* clear interrupt status */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_STATUS_REG,
	    IMX51_PWM_STATUS_FE |
	    IMX51_PWM_STATUS_ROV |
	    IMX51_PWM_STATUS_CMP |
	    IMX51_PWM_STATUS_FWE);

	/* set pulse period */
	val = IMX51_IPG_CLK / (256 + 1) / 300;
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_PERIOD_REG, val);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_CONTROL_REG,
	    IMX51_PWM_CONTROL_EN |
	    IMX51_PWM_CONTROL_REPEAT(0) |
	    IMX51_PWM_CONTROL_PRESCALER(255) |
	    IMX51_PWM_CONTROL_CLKSRC(1) | /* 1 is IPG_CLK */
	    IMX51_PWM_CONTROL_POUTC(0) |
	    IMX51_PWM_CONTROL_WAITEN |
	    IMX51_PWM_CONTROL_FWM(0));
}

static void
imx51pwm_rate(struct imx51pwm_softc *sc, int rate)
{
	uint32_t val;

	if (rate > 255)
		rate = 255;

	val = bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_PERIOD_REG);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_SAMPLE_REG,
	    (val * rate / 255) & 0xffff);
}

static int
imx51pwm_intr(void *arg)
{
	struct imx51pwm_softc *sc;

	sc = arg;

	/* DEBUG */
	printf("%s:%d: 0x%08x\n", __func__, __LINE__, bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_STATUS_REG));

	/* clear status */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, IMX51_PWM_STATUS_REG,
	    IMX51_PWM_STATUS_FE |
	    IMX51_PWM_STATUS_ROV |
	    IMX51_PWM_STATUS_CMP |
	    IMX51_PWM_STATUS_FWE);

	return 1;
}
